"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const FriendlyErrorsWebpackPlugin = require("friendly-errors-webpack-plugin");
const CopyWebpackPlugin = require("copy-webpack-plugin");
const OptimizeCssAssetsPlugin = require("optimize-css-assets-webpack-plugin");
const SpeedMeasurePlugin = require("speed-measure-webpack-plugin");
const CleanWebpackPlugin = require("clean-webpack-plugin");
const HtmlWebpackPlugin = require("html-webpack-plugin");
const MiniCssExtractPlugin = require("mini-css-extract-plugin");
const path_1 = require("path");
const webpack = require("webpack");
const webpack_bundle_analyzer_1 = require("webpack-bundle-analyzer");
const webpackMerge = require("webpack-merge");
const nodeExternals = require("webpack-node-externals");
const WebpackBar = require("webpackbar");
const TerserPlugin = require("terser-webpack-plugin");
const utils_1 = require("./utils");
const configuration = (env = "development", userConfig, isDebugging = false, port) => {
    const isDev = env !== "production";
    const context = utils_1.rootPath;
    const paths = utils_1.getPaths(userConfig.paths);
    const allConfigs = {
        rules: {
            typescript: {
                test: /\.tsx?$/,
                use: [
                    {
                        loader: require.resolve("ts-loader"),
                        options: {
                            context: utils_1.rootPath,
                            transpileOnly: true,
                            experimentalWatchApi: true
                        }
                    }
                ],
                include: paths.src,
                exclude: /node_modules/
            },
            babel: {
                test: /\.jsx?$/,
                use: [
                    {
                        loader: require.resolve("babel-loader")
                    }
                ],
                include: paths.src,
                exclude: /node_modules/
            },
            html: {
                test: /\.(html)$/,
                use: { loader: require.resolve("html-loader") },
                include: paths.src,
                exclude: /node_modules/
            },
            scss: {
                test: /\.(sa|sc|c)ss$/,
                use: [
                    isDev ? require.resolve("style-loader") : MiniCssExtractPlugin.loader,
                    {
                        loader: require.resolve("css-loader"),
                        options: { sourceMap: !isDev, importLoaders: 1 }
                    },
                    {
                        loader: require.resolve("postcss-loader"),
                        options: { sourceMap: !isDev, plugins: [require("autoprefixer")] }
                    },
                    {
                        loader: require.resolve("sass-loader"),
                        options: { sourceMap: !isDev }
                    }
                ]
            },
            images: {
                test: /\.(png|svg|jpg|gif)$/,
                loader: require.resolve("url-loader"),
                options: { limit: 10000, name: "images/[name].[hash:7].[ext]" }
            },
            fonts: {
                test: /\.(woff|woff2|eot|ttf|otf)$/,
                loader: require.resolve("url-loader"),
                options: { limit: 10000, name: "fonts/[name].[hash:7].[ext]" }
            },
            media: {
                test: /\.(mp4|webm|ogg|mp3|wav|flac|aac)(\?.*)?$/,
                loader: require.resolve("url-loader"),
                options: { limit: 10000, name: "media/[name].[hash:7].[ext]" }
            }
        },
        plugins: {
            miniCssExtract: new MiniCssExtractPlugin({
                filename: isDev ? "[name].css" : "css/[name].[contenthash].css",
                chunkFilename: isDev ? "[id].css" : "css/[id].[chunkhash].css"
            }),
            friendlyErrors: new FriendlyErrorsWebpackPlugin(),
            define: new webpack.DefinePlugin({
                "process.env": isDev
                    ? utils_1.getFile("./env.development.ts")
                    : utils_1.getFile("./env.production.ts")
            }),
            html: new HtmlWebpackPlugin({
                filename: "index.html",
                template: utils_1.resolveHtmlTemplate(),
                inject: true,
                minify: isDev
                    ? false
                    : {
                        removeComments: true,
                        collapseWhitespace: true,
                        removeAttributeQuotes: true
                    }
            }),
            clean: new CleanWebpackPlugin(["dist"], { root: utils_1.rootPath }),
            hmr: new webpack.HotModuleReplacementPlugin(),
            webpackbar: new WebpackBar({
                name: "TypePack"
            }),
            analyze: new webpack_bundle_analyzer_1.BundleAnalyzerPlugin(),
            copyStatic: new CopyWebpackPlugin([
                {
                    from: path_1.join(utils_1.rootPath, "static"),
                    to: paths.outputFolder,
                    ignore: [".*"]
                }
            ])
        }
    };
    const { rules, plugins } = Object.assign({}, allConfigs, userConfig);
    const commonConfig = {
        context,
        mode: isDev ? "development" : "production",
        devtool: isDev ? "cheap-module-eval-source-map" : "source-map",
        entry: isDev && !utils_1.isBackEndConfig(userConfig.mode)
            ? utils_1.makeDevEntries(paths.entry, port)
            : paths.entry,
        optimization: isDev
            ? {}
            : {
                removeAvailableModules: false,
                removeEmptyChunks: false,
                minimizer: [
                    new TerserPlugin({
                        cache: true,
                        parallel: true,
                        sourceMap: true
                    }),
                    new OptimizeCssAssetsPlugin({
                        cssProcessor: require("cssnano"),
                        cssProcessorPluginOptions: {
                            preset: [
                                "default",
                                {
                                    discardComments: {
                                        removeAll: true
                                    }
                                }
                            ]
                        },
                        canPrint: true
                    })
                ],
                splitChunks: {
                    chunks: "all"
                }
            },
        resolve: {
            symlinks: false,
            extensions: [".tsx", ".ts", ".js", ".jsx", ".json"],
            modules: [
                "node_modules",
                utils_1.resolveContext(context, "node_modules"),
                utils_1.resolveLocal("node_modules")
            ],
            alias: {
                "@": paths.src
            }
        },
        resolveLoader: {
            modules: [
                "node_modules",
                utils_1.resolveContext(context, "node_modules"),
                utils_1.resolveLocal("node_modules")
            ]
        },
        module: {
            rules: [rules.typescript, rules.babel]
        },
        plugins: [
            plugins.friendlyErrors,
            plugins.webpackbar,
            plugins.clean,
            plugins.define,
            ...(process.env.BUNDLE_ANALYZE ? [plugins.analyze] : [])
        ],
        output: {
            publicPath: "/",
            filename: isDev ? "js/[name].js" : "js/[name].[chunkhash].js",
            chunkFilename: isDev ? "js/[id].js" : "js/[id].[chunkhash].js",
            path: paths.outputFolder,
            libraryTarget: "umd",
            pathinfo: false
        }
    };
    const webAppConfig = webpackMerge(commonConfig, {
        module: {
            rules: [rules.html, rules.scss, rules.images, rules.fonts, rules.media]
        },
        plugins: [
            plugins.html,
            ...(isDev ? [plugins.hmr] : [plugins.miniCssExtract])
        ]
    });
    const backEndConfig = webpackMerge(commonConfig, {
        target: "node",
        externals: [nodeExternals()],
        output: {
            filename: "index.js",
            path: path_1.resolve("./dist"),
            libraryTarget: "commonjs2"
        }
    });
    let resultConfig = webAppConfig;
    if (utils_1.isBackEndConfig(userConfig.mode)) {
        resultConfig = backEndConfig;
    }
    if (typeof userConfig.config === "function") {
        userConfig.config(resultConfig);
    }
    if (isDebugging) {
        console.log(JSON.stringify(resultConfig));
    }
    const smp = new SpeedMeasurePlugin({
        disable: !process.env.SMP
    });
    return smp.wrap(resultConfig);
};
exports.default = configuration;
//# sourceMappingURL=config.js.map